' *******************************************************************************************************
' *                                                                                                     *
' *     AiChip_P4_001.spin ( INITIAL RELEASE - WORKING, BUT NOT EXHAUSTIVELY TESTED )                   *
' *                                                                                                     *
' *******************************************************************************************************

PRI Version
  return String("AiChip_P4_001")

CON
{

  Known Bugs ...

    Setting BENCHMARK=true causes crash
    
  Not implemented ..

    Opc_EQUM ( CompareWithMem )
    Opc_FLO 
    Opc_LDCN 

    Opc_CSPELN      
    Opc_CSPEOF      
    Opc_CSPGET      
    Opc_CSPNEW      
    Opc_CSPPUT      
    Opc_CSPRDC      
    Opc_CSPRDI      
    Opc_CSPRLN      
    Opc_CSPRST      
    Opc_CSPSAV      

  Test.pas ...

    program testp4( input, output, prr, prd );
    var
      i : integer;

      procedure inc;

        function add( n : integer; m : integer ) : integer;
        begin
          add := n + m;
        end;

      begin
        i := add( i, 1 );
      end;

    begin
      while( true ) do
        begin
          write( i );
          inc;
        end;
    end.
    
}
CON
{

PLEASE NOTE : If modifying and re-releasing this code, please change the prefix of the source file from
              AiChip so as to avoid any confusion as to the origin of the source and update all internal
              references as appropriate. If you wish to credit the orginal source, anything  along the
              lines of, "Based on original AiChip_P4_XXX from AiChip Industries", is acceptable.

              AiChip is a trademark of AiChip Industries.

}
CON         

' .-----------------------------------------------------------------------------------------------------.
' |     Board Configuration                                                                             |
' `-----------------------------------------------------------------------------------------------------'

  _CLKMODE            = XTAL1 + PLL16x                  ' 80MHz System Clock Speed
  _XINFREQ            = 5_000_000

  TV_PIN              = 12                              ' Base TV Pin

OBJ

  tv                  : "TV_Text"                       ' TV Text Display used for debugging

  p4                  : "Spin-P4"                       ' P4 Bytecode to execute ( Spin-P4.bin )

CON

' .-----------------------------------------------------------------------------------------------------.
' |     P4 Configuration                                                                                |
' `-----------------------------------------------------------------------------------------------------'

  BENCHMARK           = false                           ' Set true if benchmarking
  
  FORCING16           = false                           ' Set true if INT32 is forced
  FORCING32           = false                           ' Set true if INT16 is forced
  
  DEFAULT32           = true                            ' Set true if INT32 is default, otherwise INT16
  
  STACK_SIZE          = 1024                            ' Entries allowed on P4 runtime execution stack

CON

  BYTECODE_VERSION    = 0                               ' Bytecode version this VM supports

  FILE_INPUT            = p4#FILE_INPUT                 ' Input file number
  FILE_OUTPUT           = p4#FILE_OUTPUT                ' Output file number
  FILE_PRD              = p4#FILE_PRD                   ' Prd file number
  FILE_PRR              = p4#FILE_PRR                   ' Prr file number
   
  MODE_SPIN_ONLY      = %010                            ' How the P4 opcode processing can be handled
  MODE_SPIN_THEN_PASM = %011
  MODE_PASM_ONLY      = %100
  MODE_PASM_THEN_SPIN = %110
  
  HEAP_HEAD_TYPE      = 0                               ' Heap chunk Type entry offset
  HEAP_HEAD_BLEN      = 1                               ' Heap chunk ByteLen entry offset
  HEAP_HEAD_DATA      = 2                               ' Heap chunk Data entry offset

  HEAP_WORD           = 1
  HEAP_LONG           = 2
 
VAR

' .-----------------------------------------------------------------------------------------------------.
' |     P4 Runtime Data Structures                                                                      |
' `-----------------------------------------------------------------------------------------------------'

  long INT32                                            ' Base size             INT32=true  INT16=false
  long INTWORDS                                         ' Base size in words    INT32=2     INT16=1
  long INTBYTES                                         ' Base size in bytes    INT32=4     INT16=2
  
  long stack[ STACK_SIZE ]                              ' Run-time stack ( must be longs not words )

  long pcBase                                           ' Pointer to first byte of bytecode
  long spBase                                           ' Pointer to start of stack
  long npBase                                           ' Pointer to start of heap

  long pc                                               ' Current Program Counter
  long mp                                               ' Current frame Mark Pointer
  long sp                                               ' Current Stack Pointer ( poinst to TOS )
  long ep                                               ' End Pointer for stack ( highest SP will each )
  long np                                               ' New Heap allocation pointer ( last allocated )
  
  long mode                                             ' Type of P4 opcode processing
  long debugging                                        ' Set true if debugging the P4 VM
  long allowNotImplementedNatives                       ' Set true if unimplemented Native Methods ignored

  long stackAltered                                     ' True if stack has altered flag
  long opcode                                           ' Current opcode
  long rhsCached                                        ' Speed-up caching for PopL then PopL combination

  long benchmarkStack[ 100 ]                            ' Stack used by P4 VM if in own cog
    
PUB Main

  tv.start( TV_PIN )
  tv.Str( Version )
  tv.Str( String( " - " ) )
  tv.Str( p4.SourceFilename )
  tv.Str( String( ".pas" ) )
  
  if FORCING16 and FORCING32
    Crashed( String( "Cannot use FORCING16 and FORCING32" ) )

  if BYTECODE_VERSION <> p4#BYTECODE_VERSION
    tv.Str( String( $0D,"Incompatible bytecode versions",$0D ) )
    if BYTECODE_VERSION < p4#BYTECODE_VERSION
      tv.Str( String( "Upgrade ") )
      tv.Str( Version )
      Crashed( -String( ".spin") )
    else
      tv.Str( String( "Recompile " ) )
      tv.Str( p4.SourceFilename )
      Crashed( -String( ".pas" ) )

  if BENCHMARK
    CogNew( P4Start, @benchmarkStack )
    repeat
      tv.Out( $01 )
      tv.Out( $0D )
      tv.Dec( long[ $7FFC ] )
      if long[ $7FFC ] > 10_000
        tv.Str( String( " - FINISHED" ) )

  mode := MODE_SPIN_ONLY
           
  debugging := true

  P4Start
  
PRI P4Start

  P4Init
  P4Loop
  
PRI P4Init | baseSize, constPtr
  
  pcBase := p4.BytecodeStart

  baseSize := byte[ pcBase ] >> 4
  
  case baseSize
    0     : INT32 := DEFAULT32
    2     : INT32 := false
    4     : INT32 := true
    other : Crashed( String( "Invalid Base Size" ) )

  if FORCING16
    INT32 := false

  if FORCING32
    INT32 := true 

  if INT32
    INTWORDS := 2
    INTBYTES := 4
  else
    INTWORDS := 1
    INTBYTES := 2

  constPtr := byte[ pcBase ] & $F

  if constPtr
    constPtr += pcBase + 1 + 4
    PokeLong( constPtr, STACK_SIZE - 1 ) 
    constPtr += 4 + 1 + 4
    PokeLong( constPtr, STACK_SIZE - 1 )

  spBase := @stack - INTBYTES
  npBase := @stack + ( STACK_SIZE << 2 )

  if debugging
    tv.Str( String( " - INT" ) )
    tv.Dec( 16 << (INTWORDS-1) )
    
PRI P4Loop

  np := npBase

  pc := pcBase + 1
  
  mp := spBase
  sp := spBase
  ep := spBase + ( 5 << INTWORDS )
  
  repeat

    if debugging

' --- ShowStack
      Pause_Ms( 1000 )

      tv.Out( $0D )
      tv.Hex( pc, 4 )
      tv.Out( " " )
      tv.Hex( pc - pcBase, 4 )
      tv.Out( " " )

    opcode := byte[ pc++ ]

    if debugging
    
      tv.Hex( opcode, 2 )
      tv.Out( " " )
      tv.Str( p4.OpcodeNamePtr( opcode ) )

    if opcode =< p4#Opc_CSPWRS
      CspHandler
    else
      if opcode < p4#Opc_LDCI_L
        case opcode
          p4#Opc_ABI:           Push( || Pop )
          p4#Opc_ABR:           Push( FloatAbs(Pop) )
          p4#Opc_ADI:           Push( Pop + Pop )
          p4#Opc_ADR:           Push( FloatAdd(Pop,Pop) )
          p4#Opc_AND:           Push( ( Pop & Pop ) & 1 )                       ' Boolean And
          p4#Opc_CHKA_B:        ChkFromPool( "A", Operand_UnsignedByte )
          p4#Opc_CHKA_W:        ChkFromPool( "A", Operand_UnsignedWord )
          p4#Opc_CHKB:          ChkWithThis( "B", 0, 1 )
          p4#Opc_CHKC_ZB:       ChkWithThis( "C", 0, Operand_UnsignedByte )
          p4#Opc_CHKC_OB:       ChkWithThis( "C", 1, Operand_UnsignedByte )
          p4#Opc_CHKC_BB:       ChkWithThis( "C", Operand_UnsignedByte, Operand_UnsignedByte )
          p4#Opc_CHKI_B:        ChkFromPool( "I", Operand_UnsignedByte )
          p4#Opc_CHKI_W:        ChkFromPool( "I", Operand_UnsignedWord )
          p4#Opc_CHR:           Push( Pop & $FF )
          p4#Opc_CUP_ZW:        CallUp( 0, Operand_UnsignedWord )
          p4#Opc_CUP_OW:        CallUp( 1, Operand_UnsignedWord )
          p4#Opc_CUP_BW:        CallUp( Operand_UnsignedByte, Operand_UnsignedWord )
          p4#Opc_DECI_B:        Push( Pop - Operand_SignedByte )
          p4#Opc_DECI_W:        Push( Pop - Operand_SignedWord )
          p4#Opc_DECI_L:        Push( Pop - Operand_SignedLong )
          p4#Opc_DIF:           Push( Pop ^ Pop )                               ' Set difference
          p4#Opc_DVI:           Push( PopL / PopR )
          p4#Opc_DVR:           Push( FloatDiv(PopL,PopR) )
          p4#Opc_ENT1_B:        Ent1( Operand_UnsignedByte )
          p4#Opc_ENT1_W:        Ent1( Operand_UnsignedWord )
          p4#Opc_ENT2_B:        Ent2( Operand_UnsignedByte )
          p4#Opc_ENT2_W:        Ent2( Operand_UnsignedWord )
          p4#Opc_EQUX:          PushCompare( Pop == Pop )
          p4#Opc_EQUR:          PushCompare( FloatCmp(Pop,Pop) == 0 )
          p4#Opc_EQUM_W:        PushCompare( CompareWithMem(Operand_UnSignedWord) == 0 )
          p4#Opc_FJP_A:         BranchIf( Pop == 0, Operand_UnsignedWord )
          p4#Opc_FLO:       NotImplemented 
          p4#Opc_FLT:           Push( FloatInt( Pop ) )
          p4#Opc_GEQX:          PushCompare( PopL => PopR )
          p4#Opc_GEQR:          PushCompare( FloatCmp(PopL,PopR) => 0 )
          p4#Opc_GEQM_W:        PushCompare( CompareWithMem(Operand_UnSignedWord) => 0 )
          p4#Opc_GRTX:          PushCompare( PopL > PopR )
          p4#Opc_GRTR:          PushCompare( FloatCmp(PopL,PopR) > 0 )
          p4#Opc_GRTM_W:        PushCompare( CompareWithMem(Operand_UnSignedWord) > 0 )
          p4#Opc_INCX_B:        Push( Pop + Operand_SignedByte )
          p4#Opc_INCX_W:        Push( Pop + Operand_SignedWord )
          p4#Opc_INCX_L:        Push( Pop + Operand_SignedLong )
          p4#Opc_INDX_B:        Push( Peek( Pop + ( Operand_UnSignedByte << INTWORDS ) ) ) 
          p4#Opc_INDX_W:        Push( Peek( Pop + ( Operand_UnSignedWord << INTWORDS ) ) ) 
          p4#Opc_INN:           PushCompare( ( Pop & Pop ) <> 0 )               ' Set membership
          p4#Opc_INT:           Push( Pop & Pop )                               ' Set intersection
          p4#Opc_IOR:           Push( ( Pop | Pop ) & 1 )                       ' Boolean Or
          p4#Opc_IXA_B:         Push( ( Operand_UnSignedByte * Pop ) + Pop ) 
          p4#Opc_IXA_W:         Push( ( Operand_UnSignedWord * Pop ) + Pop )
          p4#Opc_LAO_B:         Push( @stack + Operand_UnSignedByte )
          p4#Opc_LAO_W:         Push( @stack + Operand_UnSignedWord ) 
          p4#Opc_LCA_B:         Push( pcBase + Operand_UnSignedByte )
          p4#Opc_LCA_W:         Push( pcBase + Operand_UnSignedWord )  
          p4#Opc_LDA_BB:        Push( mpBase(Operand_UnSignedByte) + ( Operand_UnsignedByte << INTWORDS ) )
          p4#Opc_LDA_BW:        Push( mpBase(Operand_UnSignedByte) + ( Operand_UnsignedWord << INTWORDS ) )
          p4#Opc_LDC_B:         PushSetBits( Operand_UnSignedByte )
          p4#Opc_LDC_W:         PushSetBits( Operand_UnSignedWord )
          p4#Opc_LDCB_F:        Push( 0 )
          p4#Opc_LDCB_T:        Push( 1 )
          p4#Opc_LDCC_Z:        Push( 0 )
          p4#Opc_LDCC_O:        Push( 1 )
          p4#Opc_LDCC_B:        Push( Operand_UnSignedByte )
          p4#Opc_LDCI_B:        Push( Operand_SignedByte )
          p4#Opc_LDCI_W:        Push( Operand_SignedWord )
          other:                NotImplemented
      else
        case opcode
          p4#Opc_LDCI_L:        Push( Operand_SignedLong )
          p4#Opc_LDCN:      NotImplemented 
          p4#Opc_LDOX_B:        Push( Peek( @stack + ( Operand_UnSignedByte << INTWORDS ) ) )
          p4#Opc_LDOX_W:        Push( Peek( @stack + ( Operand_UnSignedWord << INTWORDS ) ) )
          p4#Opc_LEQX:          PushCompare( PopL =< PopR )
          p4#Opc_LEQR:          PushCompare( FloatCmp(PopL,PopR) =< 0 )
          p4#Opc_LEQM_W:        PushCompare( CompareWithMem(Operand_UnSignedWord) =< 0 )
          p4#Opc_LESX:          PushCompare( PopL < PopR )
          p4#Opc_LESR:          PushCompare( FloatCmp(PopL,PopR) < 0 )
          p4#Opc_LESM_W:        PushCompare( CompareWithMem(Operand_UnSignedWord) < 0 )
          p4#Opc_LODX_BB:       Push( Peek( mpBase(Operand_UnSignedByte) + ( Operand_UnSignedByte << INTWORDS ) ) )
          p4#Opc_LODX_BW:       Push( Peek( mpBase(Operand_UnSignedByte) + ( Operand_UnSignedWord << INTWORDS ) ) )
          p4#Opc_LODA_BB:       Push( mpBase(Operand_UnSignedByte) + ( Operand_UnSignedByte << INTWORDS ) )
          p4#Opc_LODA_BW:       Push( mpBase(Operand_UnSignedByte) + ( Operand_UnSignedWord << INTWORDS ) )
          p4#Opc_MOD:           Push( PopL // PopR )
          p4#Opc_MOV_B:         Copy( PopL, PopR, Operand_UnSignedByte )
          p4#Opc_MOV_W:         Copy( PopL, PopR, Operand_UnSignedWord )
          p4#Opc_MPI:           Push( Pop * Pop )
          p4#Opc_MPR:           Push( FloatMpy(Pop,Pop) )
          p4#Opc_MST_Z:         MarkStack( 0 )
          p4#Opc_MST_O:         MarkStack( 1 )
          p4#Opc_MST_B:         MarkStack( Operand_UnSignedByte )
          p4#Opc_NEQX:          PushCompare( Pop <> Pop )
          p4#Opc_NEQR:          PushCompare( FloatCmp(Pop,Pop) <> 0 )
          p4#Opc_NEQM_W:        PushCompare( CompareWithMem(Operand_UnSignedWord) <> 0 )
          p4#Opc_NGI:           Push( - Pop )
          p4#Opc_NGR:           Push( FloatNeg(Pop) )
          p4#Opc_NOT:           Push( Pop ^ 1 )                                 ' Boolean Not
          p4#Opc_ODD:           Push( Pop & 1 )
          p4#Opc_ORDX:          ' Nop
          p4#Opc_RETX:          ReturnWithValue( true )
          p4#Opc_RETP:          ReturnWithValue( false )
          p4#Opc_SBI:           Push( PopL - PopR )
          p4#Opc_SBR:           Push( FloatSub(PopL,PopR) )
          p4#Opc_SGS:           PushSgs( Pop )                                  ' Set, Generate Single
          p4#Opc_SQI:           Push( ^^ Pop )
          p4#Opc_SQR:           Push( FloatSqr(Pop) ) 
          p4#Opc_SROX_B:        Poke( @stack + ( Operand_UnSignedByte << INTWORDS ) , Pop )                     
          p4#Opc_SROX_W:        Poke( @stack + ( Operand_UnSignedWord << INTWORDS ) , Pop )
          p4#Opc_STOX:          Poke( PopL , PopR )
          p4#Opc_STP:           Crashed( String( "Program Ended" ) )
          p4#Opc_STRX_BB:       Poke( mpBase(Operand_UnSignedByte) + ( Operand_UnSignedByte << INTWORDS ) , Pop )
          p4#Opc_STRX_BW:       poke( mpBase(Operand_UnSignedByte) + ( Operand_UnSignedWord << INTWORDS ) , Pop )             
          p4#Opc_TRC:           Push( FloatTrc(Pop) )
          p4#Opc_UJC:           Crashed( String( "Unused Jump Case" ) )
          p4#Opc_UJP_A:         BranchIf( true, Operand_UnsignedWord )
          p4#Opc_UNI:           Push( Pop | Pop )                               ' Set union / Join
          p4#Opc_XJP  :         BranchIf( true, pc + ( Pop << INTWORDS ) )
          other:                NotImplemented
     
PRI CspHandler | fileNumber, n, fieldWidth, stringSize

  case opcode
    p4#Opc_CSPELN:      NotImplemented
    p4#Opc_CSPEOF:      NotImplemented
    p4#Opc_CSPGET:      NotImplemented
    p4#Opc_CSPNEW:      NotImplemented
    p4#Opc_CSPPUT:      NotImplemented
    p4#Opc_CSPRDC:      NotImplemented
    p4#Opc_CSPRDI:      NotImplemented
    p4#Opc_CSPRLN:      NotImplemented
    p4#Opc_CSPRST:      NotImplemented
    p4#Opc_CSPSAV:      NotImplemented
    
    p4#Opc_CSPWLN:              fileNumber := Pop
                                if debugging
                                  tv.Str( String( $0D,"<CR><LF>" ) )
                                else
                                  tv.Out( $0D )

    p4#Opc_CSPWRB:              fileNumber := Pop
                                fieldWidth := Pop
                                n          := Pop
                                if debugging
                                  tv.Out( $0D )
                                if n 
                                  tv.Out( "T" )
                                else
                                  tv.Out( "F" )
                                repeat fieldWidth-1
                                  tv.Out( " " )
                                if debugging

    p4#Opc_CSPWRC:              fileNumber := Pop
                                fieldWidth := Pop
                                n          := Pop
                                if debugging
                                  tv.Out( $0D )
                                tv.Out( n )
                                repeat fieldWidth-1
                                  tv.Out( " " )
                   
    p4#Opc_CSPWRI:              fileNumber := Pop
                                fieldWidth := Pop
                                n          := Pop
                                if BENCHMARK
                                  long[ $7FFC ] := n
                                else
                                  if debugging
                                    tv.Out( $0D )
                                  tv.Dec( n )
                                
    p4#Opc_CSPWRS:              fileNumber := Pop
                                stringSize := Pop
                                fieldWidth := Pop
                                n          := Pop
                                if debugging
                                  tv.Out( $0D )
                                repeat fieldWidth
                                  if byte[ n ] 
                                    tv.Out( byte[n++] )
                                  else
                                    tv.Out( " " )

    p4#Opc_CSPWRU:              fileNumber := Pop
                                fieldWidth := Pop
                                n          := Pop
                                if debugging
                                  tv.Out( $0D )
                                tv.Bin( n, 16 << (INTWORDS-1) )

CON
 
' .-----------------------------------------------------------------------------------------------------.
' |        Flow Control                                                                                 |
' `-----------------------------------------------------------------------------------------------------'

'   Initially ...      After Mst ...      After Cup ...      After Ent 1 ...    After Ent 2 ...                 After Csp ...
'
'         .----.             .----.             .----.             .----.             .----.
'   SP -->| XX |             | XX |             | XX |             | XX |             | XX |
'         `----'             |----|             |----|             |----|             |----|
'                   (RetVal) | 00 |       MP -->| 00 |       MP -->| 00 |       MP -->| 00 |
'                            | XP |             | XP |             | XP |             | XP |
'                            | MP |             | MP |             | MP |             | MP |
'                            | EP |             | EP |             | EP |             | EP |
'                      SP -->| ?? |             | PC |             | PC |             | PC |            
'                            `----'             |----|             |----|             |----|
'                                      (Params) | P0 |             | P0 |             | P0 |
'                                               :    :             :    :             :    :
'                                         SP -->| Pn |             | Pn |             | Pn |
'                                               `----'             |----|             |----|
'                                                         (Locals) | L0 |             | L0 |
'                                                                  :    :             :    :
'                                                            SP -->| Ln |       SP -->| Ln |
'                                                                  `----'             |----|
'                                                                           (Runtime) |    |
'                                                                                     :    :
'                                                                               EP -->|    |
'                                                                                     `----'
' 1    MST     Create a stack frame
' 2    ---     Push local parameters
' 3    CUP     Set MP to where it should be, set return PC
' 4    ENT1    Allocate space for local vars
' 5    ENT2    Allocate space for runtime use of stack
'
' MP always points to the start of the frame stack which is also the place where a function return
' value is placed.
'
' XP is a link to previous MP's.
'
' There seems to be no reason that XP and PC couldn't be swapped. This would allow the RETx opcodes
' to be more efficient as they could simply start un-popping at EP and not have to skip XP which is
' thrown away later.
 
PRI Ent1( n )

  stackAltered := true
  
  sp := mp + ( n << INTWORDS )
  if sp > np
    Crashed( String( "Stack Overflow - Local Vars" ) )

PRI Ent2( n )

  ep := sp + ( n << INTWORDS )
  if ep > np
    Crashed( String( "Stack Overflow - Runtime Vars" ) )

PRI MarkStack( n )

  Push( 0 )                     ' Return Value
  Push( mpBase(n) )
  Push( mp )
  Push( ep )
  Push( 0 )                     ' Will be replaced by PC
  
PRI CallUp( n, destination )    ' n = How many locals there were

  mp := sp - ( ( n + 4 ) << INTWORDS ) 
  
  Poke( mp + ( 4 << INTWORDS ) , pc )
  
  pc := destination + pcBase

PRI ReturnWithValue( withValue )

  sp := mp + ( 4 << INTWORDS )  ' Get ready to pop the stack frame

  pc := Pop                     ' Restore PC
  ep := Pop                     ' Restore EP
  mp := Pop                     ' Restore MP
  Pop                           ' Ignore  BP
  if not withValue
    Pop                         ' Remove return value
  
PRI mpBase( n )

  result := mp
  repeat while n-- > 0
    result := Peek( result + ( 1 << INTWORDS ) )
  
PRI BranchIf( condition, destination )

  if condition
    pc := destination + pcBase
      
' .-----------------------------------------------------------------------------------------------------.
' |     Bytecode Operand Handling                                                                       |
' `-----------------------------------------------------------------------------------------------------'

PRI Operand_UnSignedLong : operand

  operand := byte[ pc++ ] << 24
  operand |= byte[ pc++ ] << 16
  operand |= byte[ pc++ ] << 8
  operand |= byte[ pc++ ]

  if debugging
    tv.Out( " " )
    tv.Hex( operand, 8 )

PRI Operand_UnSignedTrip : operand

  operand := byte[ pc++ ] << 16
  operand |= byte[ pc++ ] << 8
  operand |= byte[ pc++ ]

  if debugging
    tv.Out( " " )
    tv.Hex( operand, 6 )
    
PRI Operand_UnSignedWord : operand

  operand := byte[ pc++ ] << 8
  operand |= byte[ pc++ ]

  if debugging
    tv.Out( " " )
    tv.Hex( operand, 4 )

PRI Operand_UnSignedByte : operand

  operand := byte[ pc++ ]

  if debugging
    tv.Out( " " )
    tv.Hex( operand, 2 )

PRI Operand_SignedLong : operand

  operand := Operand_UnSignedLong 

  if debugging
    tv.Str( String( " (" ) )
    tv.dec( operand )
    tv.Out( ")" )
    
PRI Operand_SignedTrip : operand

  operand := SignExtendTrip( Operand_UnSignedTrip )

  if debugging
    tv.Str( String( " (" ) )
    tv.dec( operand )
    tv.Out( ")" )
        
PRI Operand_SignedWord : operand

  operand := SignExtendWord( Operand_UnSignedWord )

  if debugging
    tv.Str( String( " (" ) )
    tv.dec( operand )
    tv.Out( ")" )

PRI Operand_SignedByte : operand

  operand := SignExtendByte( Operand_UnSignedByte )

  if debugging
    tv.Str( String( " (" ) )
    tv.dec( operand )
    tv.Out( ")" )

PRI SignExtendTrip( n )

  result := n
  if result & $00_80_00_00
    result |= $FF_00_00_00

PRI SignExtendWord( n )

  result := ~~ n

PRI SignExtendByte( n )

  result := ~ n

' .-----------------------------------------------------------------------------------------------------.
' |        Top of Stack Bounds Checking                                                                 |
' `-----------------------------------------------------------------------------------------------------'

PRI ChkFromPool( typChar, constPtr ) | format, minValue, maxValue

  constPtr += pcBase
  
  format   := byte[ constPtr++ ]

  minValue := GetPoolConst( format >> 4 , constPtr )
  constPtr += format >> 4
  maxValue := GetPoolConst( format & $F , constPtr )

  ChkWithThis( typChar, minValue, maxValue )  

PRI ChkWithThis( typChar, minValue, maxValue )

  if Tos < minValue
    tv.Out( $0D )
    tv.Str( TypeNamePtr( typChar ) )
    Crashed( -String( " below minimum value" ) )
    
  if Tos > maxValue 
    tv.Out( $0D )
    tv.Str( TypeNamePtr( typChar ) )
    Crashed( -String( " above maximum value" ) )

PRI GetPoolConst( constSize, adr )

  if constSize & 8
    result := - GetPoolConst( constSize & 7, adr )
  else
    case constSize
      0 : result := 0
      1 : result := PeekByte( adr )
      2 : result := PeekWord( adr )
      3 : result := PeekTrip( adr )
      4 : result := PeekLong( adr )
      
' .-----------------------------------------------------------------------------------------------------.
' |        Floating point Routines                                                                      |
' `-----------------------------------------------------------------------------------------------------'

PRI FloatAdd( fLhs, fRhs )
  NotImplemented

PRI FloatSub( fLhs, fRhs )
  NotImplemented

PRI FloatMpy( fLhs, fRhs )
  NotImplemented

PRI FloatDiv( fLhs, fRhs )
  NotImplemented

PRI FloatCmp( fLhs, fRhs )      ' Returns -Ve if lhs < rhs, 0 id lhs == rhs, +Ve if lhs > rhs
  NotImplemented

PRI FloatAbs( fNum )
  NotImplemented

PRI FloatNeg( fNum )
  NotImplemented

PRI FloatSqr( fNum)
  NotImplemented
  
PRI FloatInt( n )               ' Convert integer to float
  NotImplemented

PRI FloatTrc( fNum )            ' Convert float to integer
  NotImplemented
    
' .-----------------------------------------------------------------------------------------------------.
' |        Push Stack                                                                                   |
' `-----------------------------------------------------------------------------------------------------'

PRI Push( n )

  stackAltered := true
  if INT32
    sp += 4
    long[ sp ] := n
  else
    sp += 2
    word[ sp ] := n

PRI PushCompare( n )

  Push( n & 1 )

PRI PushSgs( n )

  if n < 0
    Crashed( String( "Invalid bit number" ) )

  if n => ( 16 << (INTWORDS-1) )
    Crashed( String( "Bit number too large" ) )

  Push( |< n )
  
PRI PushSetBits( n ) | adr, byteCount, setBits

  adr := @stack + ( n << INTWORDS )

  byteCount := byte[ adr++ ]

  if byteCount > INTBYTES
    Crashed( String( "Constant of 'Set' too big" ) )

  setBits := 0
  repeat while byteCount-- > 0
    setBits := setBits << 8 | byte[ adr++ ]

  Push( setBits ) 
    
' .-----------------------------------------------------------------------------------------------------.
' |        Look at Top Of Stack but do not pop anything                                                 |
' `-----------------------------------------------------------------------------------------------------'

PRI Tos

  if INT32
    result := long[ sp ]
  else
    result := SignExtendWord( word[ sp ] )
    
' .-----------------------------------------------------------------------------------------------------.
' |        Pop Stack                                                                                    |
' `-----------------------------------------------------------------------------------------------------'

PRI Pop

  stackAltered := true
  if INT32
    result := long[ sp ]
    sp -= 4
  else
    result := SignExtendWord( word[ sp ] )
    sp -= 2

PRI PopL : lhs

   rhsCached := Pop
   lhs       := Pop

PRI PopR : rhs

  rhs := rhsCached

' .-----------------------------------------------------------------------------------------------------.
' |                                                                                                     |
' `-----------------------------------------------------------------------------------------------------'

PRI Poke( a, n )

  if INT32
    PokeLong( a, n )
  else
    PokeWord( a, n )

PRI PokeLong( a, n )

  if a & 3
    PokeWord( a + 0 , n >> 16 )
    PokeWord( a + 2 , n       )
  else
    long[ a ] := n

PRI PokeTrip( a, n )

  PokeByte( a + 0 , n >> 16 )
  PokeWord( a + 1 , n       )

PRI PokeWord( a, n )

  if a & 1
    PokeByte( a + 0 , n >> 8 )
    PokeByte( a + 1 , n      )
  else
    word[ a ] := n

PRI PokeByte( a, n )

  byte[ a ] := n
    
' .-----------------------------------------------------------------------------------------------------.
' |                                                                                                     |
' `-----------------------------------------------------------------------------------------------------'

PRI Peek( a )

  if INT32
    result := PeekLong( a )
  else
    result := PeekWord( a )

PRI PeekLong( a )

  if a & 3
    result := PeekWord( a + 0 ) << 16
    result |= PeekWord( a + 2 )
  else
    result := long[ a ]

PRI PeekTrip( a )

  result := PeekByte( a + 0 ) << 16
  result |= PeekWord( a + 1 )
   
PRI PeekWord( a )

  if a & 1
    result := PeekByte( a + 0 ) << 8
    result |= PeekByte( a + 1 )
  else
    result := word[ a ]

PRI PeekByte( a )

  result := byte[ a ]

' .-----------------------------------------------------------------------------------------------------.
' |                                                                                                     |
' `-----------------------------------------------------------------------------------------------------'

PRI Copy( toAdr, fromAdr, count )

  if INT32
    LongMove( toAdr, fromAdr, count )
  else
    WordMove( toAdr, fromAdr, count )

PRI CompareWithMem( adr )
  NotImplemented
    
' .-----------------------------------------------------------------------------------------------------.
' |    Heap Management                                                                                  |
' `-----------------------------------------------------------------------------------------------------'
'
' How an array is held in memory
' .------.--------.---.---.- - -.---.
' | Type | BytLen | 0 | 1 |     | N |
' `------^--------^---^---^- - -^---'
'
' An Array Reference is a pointer to the Type
' The Type is an INT that holds a value of jem#T_XXXX
' The BytLen is the number of bytes used to hold the array data
' A returned Array Start is a pointer to the first byte of data
' The Type and Bytlen are always Int(16) or Int(32) depending upon INT32
' The array data entity size depends upon Typ

PRI GetFromHeap( typ, len ) | bytLen, i

  bytLen := len << BytesShiftPerType(typ)

  ' tv.Out($0D)
  ' tv.Str( String( "Array Typ  " ) )
  ' tv.Dec( typ )
  ' tv.Out( " ")
  ' tv.Str( jem.TypeNamePtr(typ) )
  ' tv.Out($0D)
  ' tv.Str( String( "Array Len  " ) )
  ' tv.Dec( len )
  ' tv.Out($0D)
  ' tv.Str( String( "Array BytLen " ) )
  ' tv.Dec( bytLen )
  ' Pause_Ms( 10_000 )
  
  np &= $FFFF_FFFE
  np -= bytLen + ( HEAP_HEAD_DATA << HEAP_WORD )
  word[ np + ( HEAP_HEAD_TYPE << HEAP_WORD ) ] := typ
  word[ np + ( HEAP_HEAD_BLEN << HEAP_WORD ) ] := bytLen
  i  := np + ( HEAP_HEAD_DATA << HEAP_WORD )

  repeat bytLen
    byte[ i++ ] := 0
    
  return np

PRI GetHeapInfo( ptr, atTyp, atLen, atDat ) : nextPtr | typ, Bytlen, datPtr
 
  typ    := word[ ptr + ( HEAP_HEAD_TYPE << HEAP_WORD ) ]
  bytLen := word[ ptr + ( HEAP_HEAD_BLEN << HEAP_WORD ) ]
  datPtr := ptr + ( HEAP_HEAD_DATA << HEAP_WORD )
  nextPtr := datPtr + bytLen
  if nextPtr & 1
    nextPtr++

  if atTyp
    long[ atTyp ] := typ
  if atLen
    long[ atLen ] := bytLen
  if atdat
    long[ atDat ] := datPtr

PRI TypeNamePtr( typChar )

  case typChar & $7F

    "A"   : Result := String("Address")
    "B"   : Result := String("Boolean")
    "C"   : Result := String("Char")
    "I"   : Result := String("Integer")
    "R"   : Result := String("Real")
    "S"   : Result := String("Set")
    "?"   : Result := String("<Any>")
    other : Result := String("<Unknown>")

PRI TypeSizeInBytes( typChar )

  case typChar & $7F

    "A"   : Result := INTBYTES                          ' Address       32-bit
    "B"   : Result := 1                                 ' Boolean        8-bit  0..1
    "C"   : Result := 1                                 ' Char           8-bit  0..255
    "I"   : Result := INTBYTES                          ' Integer       32-bit
    "R"   : Result := INTBYTES                          ' Real / Float  32-bit
    "S"   : Result := INTBYTES                          ' Set / Union   32-bit
    
    other : Result := 0                                 ' Unknown

PRI TypeSizeInBits( typChar )

  result := TypeSizeInBytes(typChar) << 3 
      
PRI BytesShiftPerType( typChar )

  case typChar & $7F
  
    "A":   result := ( >| INTBYTES ) - 1                ' Address
    "B":   result := ( >| 1        ) - 1                ' Boolean
    "C":   result := ( >| 1        ) - 1                ' Char
    "I":   result := ( >| INTBYTES ) - 1                ' Integer
    "R":   result := ( >| INTBYTES ) - 1                ' Real
    "S":   result := ( >| INTBYTES ) - 1                ' Set
    
    other: Crashed( String( "Unknown Type" ) )

' .-----------------------------------------------------------------------------------------------------.
' |        Utility Code                                                                                 |
' `-----------------------------------------------------------------------------------------------------'

PRI ShowStack

  if stackAltered
    stackAltered := false
    tv.Out( $0D )
    tv.Hex( sp, 4 )
    tv.Out( " " )
    tv.Hex( ( ( sp - @stack ) >> INTWORDS ) + 1, 2)
    tv.out( " " )
    tv.out( " " )
    if  INT32
      ShowStackLong( 2 )
      ShowStackLong( 1 )
      ShowStackLong( 0 )
    else
      ShowStackWord( 5 )
      ShowStackWord( 4 )
      ShowStackWord( 3 )
      ShowStackWord( 2 )
      ShowStackWord( 1 )
      ShowStackWord( 0 )
  
PRI showStackLong( depth )

  tv.out( " " )
  if ( ( sp - @stack ) >> INTWORDS ) => depth
    tv.Hex( long[ sp - ( depth << INTWORDS ) ], 8 )
  else
    tv.Str( String("--------") )

PRI showStackWord( depth )

  tv.out( " " )
  if ( ( sp - @stack ) >> INTWORDS ) => depth
    tv.Hex( word[ sp - ( depth << INTWORDS ) ], 4 )
  else
    tv.Str( String("----") )

PRI ShowHeap | ptr, typChar, bytLen, datPtr, nextPtr

  ptr := np
  repeat while ptr < npBase
    nextPtr := GetHeapInfo( ptr, @typChar, @bytLen, 0 )
    ShowHeapEntry( ptr, typChar, bytLen )
    ptr := nextPtr
    
PRI ShowHeapEntry( ptr, typChar, bytLen )

  tv.Out( $0D )
  tv.Hex( ptr, 4 )
  tv.Out( "-" )
  tv.Hex( ptr + bytLen + ( HEAP_HEAD_DATA << HEAP_WORD ) - 1, 4 )
  tv.Out( " " )
  tv.Hex( typChar, 2 )
  tv.Out( " " )
  tv.Str( TypeNamePtr(typChar) )
  tv.Dec( TypeSizeInBits(typChar) ) 
  tv.Out( "[" )
  tv.dec( bytLen >> BytesShiftPerType(typChar) )
  tv.Out( "]" )  
  tv.Out( " " )
  tv.Dec( HEAP_HEAD_DATA << HEAP_WORD )
  tv.Out( "+" )
  tv.Dec( bytLen )
  tv.Str( String( " Bytes" ) )
  if typChar & $80
    tv.Str( String( " DEAD" ) )

PRI CompactHeap | ptr, typChar, bytLen, datPtr, nextPtr, nextTypChar
{
  This should work but not tested and never called
  
  ptr := heap
  repeat while ptr < HEAP_BASE
    nextPtr := GetHeapInfo( ptr, @typChar, @bytLen, @datPtr )
    if typChar & $80
      if ptr == heap
        heap := nextPtr
      else
        if nextPtr < HEAP_BASE
          GetHeapInfo( nextPtr, @nextTypChar, 0, 0 )
          if nextTypChar & $80
            ' Combine the two
            nextPtr := GetHeapInfo( nextPtr, 0, 0, 0 )
            if ( ChipVer == 1 ) or ( not BIGARRAYS )
              word[ ptr + ( HEAP_HEAD_BLEN << HEAP_WORD ) ] := nextPtr - datPtr 
            else
              long[ ptr + ( HEAP_HEAD_BLEN << HEAP_LONG ) ] := nextPtr - datPtr  
            nextPtr := ptr
    ptr := nextPtr
}
               
' .-----------------------------------------------------------------------------------------------------.
' |                                                                                                     |
' `-----------------------------------------------------------------------------------------------------'

PRI NotImplemented

  Crashed( String( "Not Implemented" ) )

PRI Crashed( txtPtr )

  if txtPtr
    if txtPtr > 0
      tv.Out( $0D )
    tv.Str( || txtPtr )

  repeat

PRI Pause_Ms( ms )

  waitcnt( CLKFREQ / 1000 * ms + CNT )

CON

' .-----------------------------------------------------------------------------------------------------.
' |       From PINT.PAS                                                                                 |
' `-----------------------------------------------------------------------------------------------------'
 
{
          75,76,77,78,79,
          3 (*sro*): begin  store[q] := store[sp];
                      sp := sp-1
                     end;

          4 (*lda*): begin sp := sp+1;
                      store[sp].va := base(p) + q
                     end;

          5 (*lao*): begin sp := sp+1;
                      store[sp].va := q
                     end;

          80,81,82,83,84,
          6 (*sto*): begin
                      store[store[sp-1].va] := store[sp];
                      sp := sp-2;
                     end;

          7 (*ldc*): begin sp := sp+1;
                      if p=1 then
                      begin store[sp].vi := q;
                      end else
                          if p = 6 then store[sp].vc := chr(q)
                          else
                            if p = 3 then store[sp].vb := q = 1
                            else (* load nil *) store[sp].va := maxstr
                     end;

          8 (*lci*): begin sp := sp+1;
                      store[sp] := store[q]
                     end;


          (* These were broken out according to type *)
          90,91,92,
          10 (*inc*): store[sp].vi := store[sp].vi+q;
          93 (*incb*): store[sp].vb := succ(store[sp].vb);
          94 (*incc*): store[sp].vc := chr(ord(store[sp].vc)+q);

          11 (*mst*): begin (*p=level of calling procedure minus level of called
                              procedure + 1;  set dl and sl, increment sp*)
                       (* then length of this element is
                          max(intsize,realsize,boolsize,charsize,ptrsize *)
                       store[sp+2].vm := base(p);
                       (* the length of this element is ptrsize *)
                       store[sp+3].vm := mp;
                       (* idem *)
                       store[sp+4].vm := ep;
                       (* idem *)
                       sp := sp+5
                      end;

          12 (*cup*): begin (*p=no of locations for parameters, q=entry point*)
                       mp := sp-(p+4);
                       store[mp+4].vm := pc;
                       pc := q
                      end;

          13 (*ent*): if p = 1 then
                        begin sp := mp + q; (*q = length of dataseg*)
                          if sp > np then errori(' store overflow          ');
                        end
                      else
                        begin ep := sp+q;
                          if ep > np then errori(' store overflow          ');
                        end;
                        (*q = max space required on stack*)

          14 (*ret*): begin case p of
                                 0: sp:= mp-1;
                                 1,2,3,4,5: sp:= mp
                            end;
                            pc := store[mp+4].vm;
                            ep := store[mp+3].vm;
                            mp:= store[mp+2].vm;
                      end;

          15 (*csp*): callsp;

          16 (*ixa*): begin
                       i := store[sp].vi;
                       sp := sp-1;
                       store[sp].va := q*i+store[sp].va;
                      end;

          17 (*equ*): begin  sp := sp-1;
                       case p of
                         1: store[sp].vb := store[sp].vi = store[sp+1].vi;
                         0: store[sp].vb := store[sp].va = store[sp+1].va;
                         6: store[sp].vb := store[sp].vc = store[sp+1].vc;
                         2: store[sp].vb := store[sp].vr = store[sp+1].vr;
                         3: store[sp].vb := store[sp].vb = store[sp+1].vb;
                         4: store[sp].vb := store[sp].vs = store[sp+1].vs;
                         5: begin  compare;
                               store[sp].vb := b;
                            end;
                       end; (*case p*)
                      end;

          18 (*neq*): begin  sp := sp-1;
                       case p of
                         0: store[sp].vb := store[sp].va <> store[sp+1].va;
                         1: store[sp].vb := store[sp].vi <> store[sp+1].vi;
                         6: store[sp].vb := store[sp].vc <> store[sp+1].vc;
                         2: store[sp].vb := store[sp].vr <> store[sp+1].vr;
                         3: store[sp].vb := store[sp].vb <> store[sp+1].vb;
                         4: store[sp].vb := store[sp].vs <> store[sp+1].vs;
                         5: begin  compare;
                               store[sp].vb := not b;
                            end
                       end; (*case p*)
                      end;

          19 (*geq*): begin  sp := sp-1;
                       case p of
                         0: errori(' <,<=,>,>= for address   ');
                         1: store[sp].vb := store[sp].vi >= store[sp+1].vi;
                         6: store[sp].vb := store[sp].vc >= store[sp+1].vc;
                         2: store[sp].vb := store[sp].vr >= store[sp+1].vr;
                         3: store[sp].vb := store[sp].vb >= store[sp+1].vb;
                         4: store[sp].vb := store[sp].vs >= store[sp+1].vs;
                         5: begin compare;
                              store[sp].vb := b or
                                (store[i1+i].vi >= store[i2+i].vi)
                            end
                       end; (*case p*)
                      end;

          20 (*grt*): begin  sp := sp-1;
                       case p of
                         0: errori(' <,<=,>,>= for address   ');
                         1: store[sp].vb := store[sp].vi > store[sp+1].vi;
                         6: store[sp].vb := store[sp].vc > store[sp+1].vc;
                         2: store[sp].vb := store[sp].vr > store[sp+1].vr;
                         3: store[sp].vb := store[sp].vb > store[sp+1].vb;
                         4: errori(' set inclusion           ');
                         5: begin  compare;
                              store[sp].vb := not b and
                                (store[i1+i].vi > store[i2+i].vi)
                            end
                       end; (*case p*)
                      end;

          21 (*leq*): begin  sp := sp-1;
                       case p of
                         0: errori(' <,<=,>,>= for address   ');
                         1: store[sp].vb := store[sp].vi <= store[sp+1].vi;
                         6: store[sp].vb := store[sp].vc <= store[sp+1].vc;
                         2: store[sp].vb := store[sp].vr <= store[sp+1].vr;
                         3: store[sp].vb := store[sp].vb <= store[sp+1].vb;
                         4: store[sp].vb := store[sp].vs <= store[sp+1].vs;
                         5: begin  compare;
                              store[sp].vb := b or
                                (store[i1+i].vi <= store[i2+i].vi)
                            end;
                       end; (*case p*)
                      end;

          22 (*les*): begin  sp := sp-1;
                       case p of
                         0: errori(' <,<=,>,>= for address   ');
                         1: store[sp].vb := store[sp].vi < store[sp+1].vi;
                         6: store[sp].vb := store[sp].vc < store[sp+1].vc;
                         2: store[sp].vb := store[sp].vr < store[sp+1].vr;
                         3: store[sp].vb := store[sp].vb < store[sp+1].vb;
                         5: begin  compare;
                              store[sp].vb := not b and
                                (store[i1+i].vi < store[i2+i].vi)
                            end
                       end; (*case p*)
                      end;

          23 (*ujp*): pc := q;

          24 (*fjp*): begin  if not store[sp].vb then pc := q;
                       sp := sp-1
                      end;

          25 (*xjp*): begin
                       pc := store[sp].vi + q;
                       sp := sp-1
                      end;

          95 (*chka*): if (store[sp].va < np) or
                          (store[sp].va > (maxstr-q)) then
                         errori(' bad pointer value       ');
          (* expanded these cases per type *)
          96,97,
          26 (*chk*): if (store[sp].vi < store[q-1].vi) or
                         (store[sp].vi > store[q].vi) then
                        errori(' value out of range      ');
          98 (*chkb*): if (store[sp].vb < store[q-1].vb) or
                          (store[sp].vb > store[q].vb) then
                        errori(' value out of range      ');
          99 (*chkc*): if (store[sp].vc < store[q-1].vc) or
                          (store[sp].vc > store[q].vc) then
                        errori(' value out of range      ');

          27 (*eof*): begin  i := store[sp].vi;
                       if i=inputadr then
                       begin store[sp].vb := eof(input);
                       end else errori(' code in error           ')
                      end;

          28 (*adi*): begin  sp := sp-1;
                       store[sp].vi := store[sp].vi + store[sp+1].vi
                      end;

          29 (*adr*): begin  sp := sp-1;
                       store[sp].vr := store[sp].vr + store[sp+1].vr
                      end;

          30 (*sbi*): begin sp := sp-1;
                       store[sp].vi := store[sp].vi - store[sp+1].vi
                      end;

          31 (*sbr*): begin  sp := sp-1;
                       store[sp].vr := store[sp].vr - store[sp+1].vr
                      end;

          32 (*sgs*): store[sp].vs := [store[sp].vi];

          33 (*flt*): store[sp].vr := store[sp].vi;

          34 (*flo*): store[sp-1].vr := store[sp-1].vi;

          35 (*trc*): store[sp].vi := trunc(store[sp].vr);

          36 (*ngi*): store[sp].vi := -store[sp].vi;

          37 (*ngr*): store[sp].vr := -store[sp].vr;

          38 (*sqi*): store[sp].vi := sqr(store[sp].vi);

          39 (*sqr*): store[sp].vr := sqr(store[sp].vr);

          40 (*abi*): store[sp].vi := abs(store[sp].vi);

          41 (*abr*): store[sp].vr := abs(store[sp].vr);

          42 (*not*): store[sp].vb := not store[sp].vb;

          43 (*and*): begin  sp := sp-1;
                       store[sp].vb := store[sp].vb and store[sp+1].vb
                      end;

          44 (*ior*): begin  sp := sp-1;
                       store[sp].vb := store[sp].vb or store[sp+1].vb
                      end;

          45 (*dif*): begin  sp := sp-1;
                       store[sp].vs := store[sp].vs - store[sp+1].vs
                      end;

          46 (*int*): begin  sp := sp-1;
                       store[sp].vs := store[sp].vs * store[sp+1].vs
                      end;

          47 (*uni*): begin  sp := sp-1;
                       store[sp].vs := store[sp].vs + store[sp+1].vs
                      end;

          48 (*inn*): begin
                       sp := sp - 1; i := store[sp].vi;
                       store[sp].vb := i in store[sp+1].vs;
                      end;

          49 (*mod*): begin  sp := sp-1;
                       store[sp].vi := store[sp].vi mod store[sp+1].vi
                      end;

          50 (*odd*): store[sp].vb := odd(store[sp].vi);

          51 (*mpi*): begin  sp := sp-1;
                       store[sp].vi := store[sp].vi * store[sp+1].vi
                      end;

          52 (*mpr*): begin  sp := sp-1;
                       store[sp].vr := store[sp].vr * store[sp+1].vr
                      end;

          53 (*dvi*): begin  sp := sp-1;
                       store[sp].vi := store[sp].vi div store[sp+1].vi
                      end;

          54 (*dvr*): begin  sp := sp-1;
                       store[sp].vr := store[sp].vr / store[sp+1].vr
                      end;

          55 (*mov*): begin i1 := store[sp-1].va;
                       i2 := store[sp].va; sp := sp-2;
                       for i3 := 0 to q-1 do store[i1+i3] := store[i2+i3]
                       (* q is a number of storage units *)
                      end;

          56 (*lca*): begin  sp := sp+1;
                       store[sp].va := q;
                      end;

 }
